# import numpy as np

# from popidd_io._widget import (
#     ExampleQWidget,
#     ImageThreshold,
#     threshold_autogenerate_widget,
#     threshold_magic_widget,
# )


# def test_threshold_autogenerate_widget():
#     # because our "widget" is a pure function, we can call it and
#     # test it independently of napari
#     im_data = np.random.random((100, 100))
#     thresholded = threshold_autogenerate_widget(im_data, 0.5)
#     assert thresholded.shape == im_data.shape
#     # etc.


# # make_napari_viewer is a pytest fixture that returns a napari viewer object
# # you don't need to import it, as long as napari is installed
# # in your testing environment
# def test_threshold_magic_widget(make_napari_viewer):
#     viewer = make_napari_viewer()
#     layer = viewer.add_image(np.random.random((100, 100)))

#     # our widget will be a MagicFactory or FunctionGui instance
#     my_widget = threshold_magic_widget()

#     # if we "call" this object, it'll execute our function
#     thresholded = my_widget(viewer.layers[0], 0.5)
#     assert thresholded.shape == layer.data.shape
#     # etc.


# def test_image_threshold_widget(make_napari_viewer):
#     viewer = make_napari_viewer()
#     layer = viewer.add_image(np.random.random((100, 100)))
#     my_widget = ImageThreshold(viewer)

#     # because we saved our widgets as attributes of the container
#     # we can set their values without having to "interact" with the viewer
#     my_widget._image_layer_combo.value = layer
#     my_widget._threshold_slider.value = 0.5

#     # this allows us to run our functions directly and ensure
#     # correct results
#     my_widget._threshold_im()
#     assert len(viewer.layers) == 2


# # capsys is a pytest fixture that captures stdout and stderr output streams
# def test_example_q_widget(make_napari_viewer, capsys):
#     # make viewer and add an image layer using our fixture
#     viewer = make_napari_viewer()
#     viewer.add_image(np.random.random((100, 100)))

#     # create our widget, passing in the viewer
#     my_widget = ExampleQWidget(viewer)

#     # call our widget method
#     my_widget._on_click()

#     # read captured output and check that it's as we expected
#     captured = capsys.readouterr()
#     assert captured.out == "napari has 1 layers\n"



# For now I will only add here support for geojson and parquet writing of shapes/points/labels layers

import pathlib

from napari.types import FullLayerData

from ._anno import point2feat, shape2feat, write_geojson


def save_manyannolayers(
    out_path: str | pathlib.Path, layers: list[FullLayerData]
) -> list[str]:
    feature_list = []
    saved_annotations = []
    for layer_data_tuple in layers:
        if layer_data_tuple[2] == "shapes":
            print("Shape layer")
            feature = shape2feat(layer_data_tuple)
            saved_annotations.append(layer_data_tuple[1]["name"])
            feature_list.append(feature)
        elif layer_data_tuple[2] == "points":
            print("Points layer")
            feature = point2feat(layer_data_tuple)
            saved_annotations.append(layer_data_tuple[1]["name"])
            feature_list.append(feature)
        else:
            print("This layer type IS NOT supported!")

    annotations = {
        "type": "FeatureCollection",
        "features": feature_list,
        "properties": {"shape_layers": saved_annotations, "prop2": "val2"},
    }

    write_geojson(out_path, annotations)

    return [str(out_path)]


# def save_geojson(out_path: str | pathlib.Path, shapes: list[FullLayerData]) -> list[str]:

#         saved_annotations = []
#         features = []
#         for layer in shapes:
#             print(len(layer))
#             print(type(layer))
#             print(layer[0])
#             print(layer[1]["name"])
#             print(layer[1].keys())
#             print(layer[2])
#             if len(layer[0]) == 1:
#                 if "qupath_comp" in layer[1].keys():
#                     shape = numpy.dot(layer[0][0], [[0, 1], [1, 0]]) # QUpath rotates and flips images.
#                 else:
#                     shape = layer[0][0]
#                 feat_geom = Polygon(shape)
#             elif len(layer[0]) > 1:
#                 geoms = []
#                 for shape in layer[0]:
#                     if "qupath_comp" in layer[1].keys():
#                         shape = numpy.dot(shape, [[0, 1], [1, 0]]) # QUpath rotates and flips images.
#                     geoms.append(Polygon(shape))
#                 feat_geom = MultiPolygon(geoms)
#             else:
#                 print("WARNING: This shape layer is empty!")
#                 continue
#             feature = {
#                 "type": "Feature",
#                 "properties": {
#                     "objectType":"annotation",
#                     "name":layer[1]["name"]},
#                     "geometry": feat_geom.__geo_interface__
#             }
#             features.append(feature)
#             saved_annotations.append(layer[1]["name"])
#         feature_collection = {
#             "type":"FeatureCollection",
#             "features": features,
#             "properties": {
#                 "shape_layers": saved_annotations,
#                 "prop2": "val2"
#             }
#         }
#         print(saved_annotations)
#         print(out_path)
#         print(f"anno_{"-".join(saved_annotations)}.geojson")
#         gdf = geopandas.GeoDataFrame.from_features(feature_collection)
#         # gdf.to_file(filename= pathlib.Path(out_path) / f"anno_{"-".join(saved_annotations)}.geojson", driver="GeoJSON")
#         gdf.to_file(filename= f"{pathlib.Path(out_path)}.geojson", driver="GeoJSON")

#         return [out_path]



import pytest

from popidd_io import get_anno_reader, get_image_reader

# make_napari_viewer is a pytest fixture that returns a napari viewer object
# you don't need to import it, as long as napari is installed
# in your testing environment

# tmp_path is a pytest fixture

# def test_image_reader(tmp_path):
#     """An example of how you might test your plugin."""

#     # write some fake data using your supported file format
#     my_test_file = str(tmp_path / "myfile.npy")
#     original_data = np.random.rand(20, 20)
#     np.save(my_test_file, original_data)

#     # try to read it back in
#     reader = napari_get_reader(my_test_file)
#     assert callable(reader)

#     # make sure we're delivering the right format
#     layer_data_list = reader(my_test_file)
#     assert isinstance(layer_data_list, list) and len(layer_data_list) > 0
#     layer_data_tuple = layer_data_list[0]
#     assert isinstance(layer_data_tuple, tuple) and len(layer_data_tuple) > 0

#     # make sure it's the same as it started
#     np.testing.assert_allclose(original_data, layer_data_tuple[0])


# def test_get_reader_pass():
#     reader = napari_get_reader("fake.file")
#     assert reader is None


# Testing Images #


@pytest.fixture
def write_im2file(tmp_path):
    def write_tiff(file):
        import numpy
        from tifffile import TiffWriter

        # TIff generation snippet from tiffile
        data = numpy.random.randint(0, 255, (8, 2, 512, 512, 3), "uint8")
        subresolutions = 2
        pixelsize = 0.29  # micrometer
        tiff_file = f"{tmp_path}/{file}"
        with TiffWriter(tiff_file, bigtiff=True) as tif:
            metadata = {
                "axes": "TCYXS",
                "SignificantBits": 8,
                "TimeIncrement": 0.1,
                "TimeIncrementUnit": "s",
                "PhysicalSizeX": pixelsize,
                "PhysicalSizeXUnit": "µm",
                "PhysicalSizeY": pixelsize,
                "PhysicalSizeYUnit": "µm",
                "Channel": {"Name": ["Channel 1", "Channel 2"]},
                "Plane": {
                    "PositionX": [0.0] * 16,
                    "PositionXUnit": ["µm"] * 16,
                },
                "Description": "A multi-dimensional, multi-resolution image",
                "MapAnnotation": {  # for OMERO
                    "Namespace": "openmicroscopy.org/PyramidResolution",
                    "1": "256 256",
                    "2": "128 128",
                },
            }
            options = {
                "photometric": "rgb",
                "tile": (128, 128),
                "compression": "jpeg",
                "resolutionunit": "CENTIMETER",
                "maxworkers": 2,
            }
            tif.write(
                data,
                subifds=subresolutions,
                resolution=(1e4 / pixelsize, 1e4 / pixelsize),
                metadata=metadata,
                **options,
            )
            # write pyramid levels to the two subifds
            # in production use resampling to generate sub-resolution images
            for level in range(subresolutions):
                mag = 2 ** (level + 1)
                tif.write(
                    data[..., ::mag, ::mag, :],
                    subfiletype=1,
                    resolution=(1e4 / mag / pixelsize, 1e4 / mag / pixelsize),
                    **options,
                )
            # add a thumbnail image as a separate series
            # it is recognized by QuPath as an associated image
            thumbnail = (data[0, 0, ::8, ::8] >> 2).astype("uint8")
            tif.write(thumbnail, metadata={"Name": "thumbnail"})

        return data, tiff_file

    return write_tiff


def test_img_reader_pass():
    reader = get_image_reader("fake.file")
    assert reader is None


def test_img_reader(write_im2file):
    data, test_file = write_im2file("temp.ome.tif")
    reader = get_image_reader(test_file)
    assert callable(reader)

    layer_data_list = reader(test_file)
    assert isinstance(layer_data_list, list) and len(layer_data_list) > 0
    assert (
        isinstance(layer_data_list[0], tuple) and len(layer_data_list[0]) > 0
    )
    assert data.shape == layer_data_list[0][0][0].shape
    # numpy.testing.assert_allclose(data, layer_data_list[0][0][0].compute()) #Numpy of 1st lvl of dask array list


def test_img_reader_list(write_im2file):
    _, path1 = write_im2file("temp1.ome.tif")
    _, path2 = write_im2file("temp2.ome.tif")
    reader = get_image_reader([path1, path2])
    assert callable(reader)


# Testing Annotations #


@pytest.fixture
def geojson_file(tmp_path):
    import geopandas as gpd
    from shapely.geometry import MultiPoint, MultiPolygon, Point, Polygon

    poly = Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])
    poly2 = Polygon([(2, 2), (3, 2), (3, 3), (2, 3)])
    multipoly = MultiPolygon([poly, poly2])
    pt = Point(0.5, 0.5)
    mpt = MultiPoint([Point(2, 2), Point(2.5, 2.5)])

    gdf = gpd.GeoDataFrame(
        {
            "name": ["poly", "multipoly", "pt", "mpt"],
            "geometry": [poly, multipoly, pt, mpt],
        }
    )

    path = tmp_path / "test_annotations.geojson"
    gdf.to_file(path, driver="GeoJSON")
    return str(path), gdf


@pytest.fixture
def geojson_pair(tmp_path):
    import geopandas as gpd
    from shapely.geometry import MultiPolygon, Polygon

    poly = Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])
    poly2 = Polygon([(2, 2), (3, 2), (3, 3), (2, 3)])
    multipoly = MultiPolygon([poly, poly2])

    gdf = gpd.GeoDataFrame(
        {"name": ["one", "two"], "geometry": [poly, multipoly]}
    )

    p1 = tmp_path / "a.geojson"
    p2 = tmp_path / "b.geojson"
    gdf.to_file(p1, driver="GeoJSON")
    gdf.to_file(p2, driver="GeoJSON")
    return [str(p1), str(p2)]


# @pytest.fixture
# def parquet_file(tmp_path):
#     import pandas as pd
#     import numpy as np

#     df = pd.DataFrame({"x": np.arange(3), "label": ["a", "b", "c"]})
#     path = tmp_path / "test_labels.parquet"
#     df.to_parquet(path)
#     return str(path), df


def test_anno_reader_pass():
    reader = get_anno_reader("fake.file")
    assert reader is None


def test_anno_reader_geojson(geojson_file):
    path, gdf = geojson_file
    reader = get_anno_reader(path)
    assert callable(reader)
    layers = reader(path)
    assert isinstance(layers, list) and len(layers) == len(gdf)
    for layer in layers:
        assert isinstance(layer, tuple)
        assert layer[2] in ("shapes", "points")


def test_anno_reader_list(geojson_pair):
    reader = get_anno_reader(geojson_pair)
    assert callable(reader)


# def test_anno_reader_parquet(parquet_file):
#     path, df = parquet_file
#     reader = get_anno_reader(path)
#     assert callable(reader)
#     layers = reader(path)
#     assert isinstance(layers, list) and len(layers) > 0
#     data_array, meta, layer_type = layers[0]
#     assert layer_type == "labels"
#     import numpy as _np
#     _np.testing.assert_array_equal(data_array, df.to_numpy())



import pathlib
import warnings
from typing import Optional
from xml.etree import ElementTree

import numpy
import tifffile
import zarr
from dask import array as darray
from napari.types import FullLayerData
from napari.utils import Colormap
from napari.utils.notifications import WarningNotification

# from typing import TYPE_CHECKING
# if TYPE_CHECKING:
#     import napari


def load_img(
    path: str | pathlib.Path,
    modality: Optional[str] = None,
    load_mem: bool = False,
) -> list[FullLayerData]:
    """
    Load an image file and convert it to a list of image layers for use in napari.

    Parameters:
    path (str | pathlib.Path): The path to the image file.
    modality (Optional[str]): The modality of the image (e.g., "BF" for Brightfield,
        "IF" for Immunofluorescence). If None, the modality will be inferred.
    load_mem (bool): Whether to load the image into memory.

    Returns:
    list[napari.types.FullLayerData]: A list of FullLayerData containing the
        image layer information.
    """

    if isinstance(path, str):
        path = pathlib.Path(path)

    zarray, int_scale, _ = read_img(path, load_mem)  # return modality
    if modality is None:
        modality = _
    else:
        print(modality)
        print(_)
        if modality == _:
            print("Selected modality matches expectations")
        else:
            print("Something is off!")
    md = read_md(path, modality)
    md["int_scale"] = int_scale

    print("Image was initiliased")

    if modality == "BF":
        print("Brightfield")
        img_layer_data = [
            (
                zarray,
                {
                    "name": path.stem,
                    "scale": md["res_scale"],
                    "metadata": md,
                    "contrast_limits": [0, md["int_scale"]],
                },
                "image",
            )
        ]
    else:
        print("Fluorescence")
        img_layer_data = []
        for index, (target, col) in enumerate(md["colmap_channels"].items()):
            md["dye"] = target
            layer = md["dye"]
            if md["new_format"] == True:
                md["biomarker"] = md["fluor_to_marker"][target]
                layer = md["biomarker"]
            cmap = Colormap(
                colors=numpy.array([(0, 0, 0, 1), col + (1,)]), name=target
            )

            channel_zarray = [array[index, :, :] for array in zarray]

            img_layer_data.append(
                (
                    channel_zarray,
                    {
                        "name": f"{layer}_{path.stem}",
                        "scale": md["res_scale"],
                        "colormap": cmap,
                        "blending": "additive",
                        "metadata": md,
                        "contrast_limits": [0, md["int_scale"]],
                    },
                    "image",
                )
            )

    return img_layer_data


def read_img(img, load_mem):
    """
    Reads an image from a given pathlib path and processes it.

    Parameters:
    img (pathlib.Path): The path to the image file.
    load_mem (bool): Flag to determine if the image should be loaded into memory.

    Returns:
    tuple: A tuple containing:
        - zarray (list): A list of zarr arrays representing the image data.
        - int_scale (int): The intensity scale of the image based on its bit depth.
        - modality (str): The modality of the image, either "BF" (Bright Field) or "IF" (Immunofluorescence).

    Raises:
    NotImplementedError: If the image bit depth is not supported.
    """

    # Loading Image data
    store = tifffile.imread(img, aszarr=True)
    image = zarr.open(store, mode="r")
    if isinstance(image, zarr.hierarchy.Group):
        zarray = [array for _, array in image.arrays()]
    else:
        zarray = [image]
    print(len(zarray), zarray[0], zarray[0].shape)
    dask_array = darray.from_zarr(
        zarray[-1]
    )  # Convert zarr array to Dask array
    max_val, mean_val = darray.compute(dask_array.max(), dask_array.mean())
    # need to add here test for image modality
    print(dask_array.shape, dask_array)
    print(mean_val)
    modality = "BF"
    if mean_val < 100:
        modality = "IF"
    if 1 < max_val <= 255:  # Image in 8bit col
        print("8bit!")
        int_scale = 255
    elif (
        255 < max_val <= 4095
    ):  # Image is 12bit colour, not 16! -> Checking actual values since Bits entry in XML lies and says 12bits for the Luca 8bit image
        print("12bit!")
        int_scale = 4095
        # int_scale = 255
    elif 4095 < max_val <= 65535:  # image in 16bit colour
        print("16 bit!")
        int_scale = 65535
        # int_scale = 255
    else:
        raise NotImplementedError("The image bit depth is not supported.")
    if not load_mem:
        zarray = [darray.from_zarr(array) for array in zarray]
    return zarray, int_scale, modality  # returns list of zarr arrays


def read_md(img, modality):
    """
    Reads metadata from a TIFF image file and returns it as a dictionary.

    Parameters:
    img (str): The file path to the TIFF image.
    modality (str): The imaging modality, e.g., "IF" (Immunofluorescence).

    Returns:
    dict: A dictionary containing the image metadata, including:
        - 'ImagePath': The file path to the image.
        - 'ResolutionUnit': The unit of resolution (e.g., centimeters or inches).
        - 'XResolution': The X resolution of the image.
        - 'YResolution': The Y resolution of the image.
        - 'res_scale': A tuple containing the scaling factors for the X and Y axes.
        - 'path': The file path to the image.
        - 'modality': The imaging modality.
        - 'fluor_to_marker' (if modality is "IF"): A mapping of fluorophores to markers.
        - 'colmap_channels' (if modality is "IF"): Color map channels.
        - 'new_format' (if modality is "IF"): A flag indicating if the new format is used.

    Raises:
    NotImplementedError: If the resolution unit is not in centimeters or inches.
    Warning: If the resolution metadata cannot be detected.
    """
    # Loading Image metadata
    res_scale = (1, 1)
    with tifffile.TiffFile(img) as src:
        full_res_tags = {
            tag.name: tag.value
            for tag in src.series[0].pages[0].tags.values()
            if tag.name
            not in [
                "StripOffsets",
                "StripByteCounts",
                "TileOffsets",
                "TileByteCounts",
            ]
            and not isinstance(tag.value, numpy.ndarray)
        }
        full_res_tags["ImagePath"] = img
        try:
            if str(full_res_tags["ResolutionUnit"]) in [
                "RESUNIT.CENTIMETER",
                "3",
            ]:
                xres = full_res_tags["XResolution"]
                yres = full_res_tags["YResolution"]
                x_scale = xres[1] / xres[0]
                y_scale = yres[1] / yres[0]
                res_scale = (x_scale, y_scale)
            elif str(full_res_tags["ResolutionUnit"]) in ["RESUNIT.INCH", "2"]:
                xres = full_res_tags["XResolution"]
                yres = full_res_tags["YResolution"]
                x_scale = (xres[1] / xres[0]) * 2.54
                y_scale = (yres[1] / yres[0]) * 2.54
                res_scale = (x_scale, y_scale)
            else:
                print(str(full_res_tags["ResolutionUnit"]))
                raise NotImplementedError(
                    "Scaling supports only images in centimeters or inches"
                )
        except Exception as e:
            warning_noMD = warnings.warn(
                f"Could not detect resolution metadata for image \n {img.stem}. \n Exception: {e}"
            )
            WarningNotification(warning_noMD)
        md = full_res_tags
        md["path"] = img
        md["modality"] = modality
        md["res_scale"] = res_scale
        if modality == "IF":
            fluor_to_marker, colmap_channels, new_format = _get_mdIF(src)
            md["fluor_to_marker"] = fluor_to_marker
            md["colmap_channels"] = colmap_channels
            md["new_format"] = new_format
    return md


def _get_mdIF(TiffFile):
    """
    Extract metadata and color mapping from a TIFF file.

    Args:
        TiffFile (TiffFile): An instance of a TIFF file object from which metadata
            and color mapping will be extracted.

    Returns:
        tuple: A tuple containing:
            - fluor_to_marker (dict or bool): A dictionary mapping fluorophores
                to markers if the new format is detected, otherwise False.
            - colmap_channels (dict): A dictionary mapping channel names to their
                respective RGB color tuples.
            - new_format (bool): A boolean indicating whether the new format was detected.
    """
    new_format = False
    single_page_md = False
    fluor_to_marker = False
    xml = ElementTree.fromstring(
        TiffFile.series[0].pages[0].tags["ImageDescription"].value
    )
    if xml.find(".//LibraryAsJSON") is not None:
        import json

        dicti = json.loads(xml.find(".//LibraryAsJSON").text)
        if "spectra" in dicti:
            new_format = True
            fluor_to_marker = {
                item["fluor"]: item["marker"]
                for item in dicti["spectra"]
                if "fluor" in item and "marker" in item
            }
    colmap_channels = {}
    for page in TiffFile.series[0].pages:
        try:
            xml = ElementTree.fromstring(page.tags["ImageDescription"].value)
            if new_format == True:
                colmap_channels[xml.find(".//Responsivity/Band/Name").text] = (
                    tuple(
                        float(x) for x in xml.find(".//Color").text.split(",")
                    )
                )
            else:
                colmap_channels[
                    xml.find(".//Responsivity/Filter/Name").text
                ] = tuple(
                    float(x) for x in xml.find(".//Color").text.split(",")
                )
        except:  # attribute error on the else above and keyerror on the imagedescription
            single_page_md = True
    # EXPERIMENTAL, to be worked upon
    if single_page_md:
        tree = ElementTree.ElementTree(xml)
        filename = f"page_{str(page)}_xml.xml"
        tree.write(filename, encoding="utf-8", xml_declaration=True)
        channels = xml.find(".//channels")
        for channel in channels.findall("channel"):
            channel_id = channel.get("id")
            channel_name = channel.get("name")
            print(channel_name)
            rgb_value = int(channel.get("rgb"))
            # Convert the RGB integer to a tuple of (R, G, B)
            r = (rgb_value >> 16) & 0xFF
            g = (rgb_value >> 8) & 0xFF
            b = rgb_value & 0xFF
            colmap_channels[channel_name] = (r, g, b)
        print(channels)
        print(channels.text)
        print(channels.text.split(","))
    print(colmap_channels)
    colmap_channels = {
        key: tuple(
            v / 255 if max(colmap_channels[key]) > 1 else v for v in val
        )
        for key, val in colmap_channels.items()
    }
    print(colmap_channels)
    return fluor_to_marker, colmap_channels, new_format



__version__ = "0.0.4"

from ._reader import get_anno_reader, get_image_reader
from ._widget import wLoadAnno, wLoadImage
from ._writer import save_manyannolayers

__all__ = (
    "get_image_reader",
    "get_anno_reader",
    "save_manyannolayers",
    "wLoadImage",
    "wLoadAnno",
)



"""
This module contains four napari widgets declared in
different ways:

- a pure Python function flagged with `autogenerate: true`
    in the plugin manifest. Type annotations are used by
    magicgui to generate widgets for each parameter. Best
    suited for simple processing tasks - usually taking
    in and/or returning a layer.
- a `magic_factory` decorated function. The `magic_factory`
    decorator allows us to customize aspects of the resulting
    GUI, including the widgets associated with each parameter.
    Best used when you have a very simple processing task,
    but want some control over the autogenerated widgets. If you
    find yourself needing to define lots of nested functions to achieve
    your functionality, maybe look at the `Container` widget!
- a `magicgui.widgets.Container` subclass. This provides lots
    of flexibility and customization options while still supporting
    `magicgui` widgets and convenience methods for creating widgets
    from type annotations. If you want to customize your widgets and
    connect callbacks, this is the best widget option for you.
- a `QWidget` subclass. This provides maximal flexibility but requires
    full specification of widget layouts, callbacks, events, etc.

References:
- Widget specification: https://napari.org/stable/plugins/guides.html?#widgets
- magicgui docs: https://pyapp-kit.github.io/magicgui/

Replace code below according to your needs.
"""

import warnings
from pathlib import Path
from typing import TYPE_CHECKING

import napari.layers
import numpy
from magicgui import magic_factory
from napari.utils.notifications import WarningNotification

from ._anno import load_geojson
from ._image import load_img

if TYPE_CHECKING:
    import napari


def image_reader(
    viewer: "napari.Viewer", bf_imgs=Path(""), if_imgs=Path(""), load_mem=bool
):
    empty_flag = True
    for img in bf_imgs:
        if img.is_file():
            img_layer_data = load_img(img, modality="BF", load_mem=load_mem)
            for (
                i
            ) in (
                img_layer_data
            ):  # unpacking list of tuples even if BF images should only have 1 layer per image
                viewer.add_layer(
                    napari.layers.Layer.create(*i)
                )  # unpack tuple as func uses positional args
                # viewer._add_layer_from_data(*i) #use this one if channel_axis present
            empty_flag = False
    for img in if_imgs:
        if img.is_file():
            img_layer_data = load_img(img, modality="IF", load_mem=load_mem)
            for i in img_layer_data:
                viewer.add_layer(napari.layers.Layer.create(*i))
            empty_flag = False
    if (
        empty_flag is True
    ):  # Should the selection be empty it will return a warning on the GUI
        warning_empty = warnings.warn(
            "No image(s) selected for loading.", stacklevel=2
        )
        WarningNotification(warning_empty)


# Test magic factory usage directly (not as decorator)
wLoadImage = magic_factory(
    function=image_reader,
    bf_imgs={
        "label": "Brightfield image(s)",
        "widget_type": "FileEdit",
        "mode": "rm",
        "filter": "*.tiff;*.tif;*.svs;*.ndpi",
    },
    if_imgs={
        "label": "Fluorescence image(s)",
        "widget_type": "FileEdit",
        "mode": "rm",
        "filter": "*.tiff;*.tif;*qptiff",
    },
    load_mem={
        "widget_type": "CheckBox",
        "value": False,
        "text": "Load full image(s) into memory",
    },
    call_button="Load image(s)",
)


def anno_reader(  # TODO: Modify this to add a checklist to confim using scale from selected image layer. Also need to add a checklist for inverting image, and add arg to load_geojson
    viewer: "napari.Viewer",
    image: "napari.layers.Image",
    anno_paths=Path(""),
):
    for anno in anno_paths:
        shape_layer_data = load_geojson(anno)
        for i in shape_layer_data:
            i[1]["scale"] = image.scale * numpy.array([-1, 1])
            viewer.add_layer(napari.layers.Layer.create(*i))


wLoadAnno = magic_factory(
    function=anno_reader,
    image={"label": "Image layer"},
    anno_paths={
        "label": "Annotation GEOJSON",
        "widget_type": "FileEdit",
        "mode": "rm",
        "filter": "*.geojson",
    },
    call_button="Load Annotation",
)


# from qtpy.QtWidgets import QHBoxLayout, QPushButton, QWidget
# from magicgui.widgets import CheckBox, Container, create_widget
# from skimage.util import img_as_float

# # Uses the `autogenerate: true` flag in the plugin manifest
# # to indicate it should be wrapped as a magicgui to autogenerate
# # a widget.
# def threshold_autogenerate_widget(
#     img: "napari.types.ImageData",
#     threshold: "float",
# ) -> "napari.types.LabelsData":
#     return img_as_float(img) > threshold


# # the magic_factory decorator lets us customize aspects of our widget
# # we specify a widget type for the threshold parameter
# # and use auto_call=True so the function is called whenever
# # the value of a parameter changes
# @magic_factory(
#     threshold={"widget_type": "FloatSlider", "max": 1}, auto_call=True
# )
# def threshold_magic_widget(
#     img_layer: "napari.layers.Image", threshold: "float"
# ) -> "napari.types.LabelsData":
#     return img_as_float(img_layer.data) > threshold

# # if we want even more control over our widget, we can use
# # magicgui `Container`
# class ImageThreshold(Container):
#     def __init__(self, viewer: "napari.viewer.Viewer"):
#         super().__init__()
#         self._viewer = viewer
#         # use create_widget to generate widgets from type annotations
#         self._image_layer_combo = create_widget(
#             label="Image", annotation="napari.layers.Image"
#         )
#         self._threshold_slider = create_widget(
#             label="Threshold", annotation=float, widget_type="FloatSlider"
#         )
#         self._threshold_slider.min = 0
#         self._threshold_slider.max = 1
#         # use magicgui widgets directly
#         self._invert_checkbox = CheckBox(text="Keep pixels below threshold")

#         # connect your own callbacks
#         self._threshold_slider.changed.connect(self._threshold_im)
#         self._invert_checkbox.changed.connect(self._threshold_im)

#         # append into/extend the container with your widgets
#         self.extend(
#             [
#                 self._image_layer_combo,
#                 self._threshold_slider,
#                 self._invert_checkbox,
#             ]
#         )

#     def _threshold_im(self):
#         image_layer = self._image_layer_combo.value
#         if image_layer is None:
#             return

#         image = img_as_float(image_layer.data)
#         name = image_layer.name + "_thresholded"
#         threshold = self._threshold_slider.value
#         if self._invert_checkbox.value:
#             thresholded = image < threshold
#         else:
#             thresholded = image > threshold
#         if name in self._viewer.layers:
#             self._viewer.layers[name].data = thresholded
#         else:
#             self._viewer.add_labels(thresholded, name=name)


# class ExampleQWidget(QWidget):
#     # your QWidget.__init__ can optionally request the napari viewer instance
#     # use a type annotation of 'napari.viewer.Viewer' for any parameter
#     def __init__(self, viewer: "napari.viewer.Viewer"):
#         super().__init__()
#         self.viewer = viewer

#         btn = QPushButton("Click me!")
#         btn.clicked.connect(self._on_click)

#         self.setLayout(QHBoxLayout())
#         self.layout().addWidget(btn)

#     def _on_click(self):
#         print("napari has", len(self.viewer.layers), "layers")



import pathlib

import geopandas
import numpy
import pandas
from napari.types import FullLayerData
from shapely import MultiPolygon, Polygon

# Add here the geojson  and parquet reading and writing support for qupath compatibility
# geojson to shapes and shapes to geojson
# parquet to shapes for cell annotations


def shape2feat(layer: FullLayerData) -> dict | None:
    if len(layer[0]) == 1:
        if "qupath_comp" in layer[1]:
            shape = numpy.dot(
                layer[0][0], [[0, 1], [1, 0]]
            )  # QUpath rotates and flips images.
        else:
            shape = layer[0][0]
        feat_geom = Polygon(shape)
    elif len(layer[0]) > 1:
        geoms = []
        for shape in layer[0]:
            if "qupath_comp" in layer[1]:
                shape = numpy.dot(
                    shape, [[0, 1], [1, 0]]
                )  # QUpath rotates and flips images.
            geoms.append(Polygon(shape))
        feat_geom = MultiPolygon(geoms)
    else:
        print("WARNING: This shape layer is empty!")
        return None
    feature = {
        "type": "Feature",
        "properties": {"objectType": "annotation", "name": layer[1]["name"]},
        "geometry": feat_geom.__geo_interface__,
    }

    return feature


def point2feat(layer: FullLayerData) -> dict:
    raise NotImplementedError
    return {}


# GeoJSON#


def load_geojson(path: str | pathlib.Path) -> list[FullLayerData]:
    """
    Load a GeoJSON file and convert it to a list of shape layers.

    Parameters:
    path (str | pathlib.Path): The path to the GeoJSON file.

    Returns:
    list[FullLayerData]: A list of FullLayerData containing the shape layer information.
    """

    if isinstance(path, str):
        path = pathlib.Path(path)

    geo_anno = geopandas.read_file(path)
    # Need to grab additional feature properties and save them as shape layer metadata
    napari_layers = []
    # Split off code below to its own function(s?) -> parse geopandas file and 1)populate annotation list, 2) generate shape_layer_data entry (or similar, need to check latest napari devs)

    for anno in geo_anno.itertuples():
        if anno.name is None:
            try:
                import ast

                name = ast.literal_eval(anno.classification)["name"]
            except (AttributeError, KeyError):
                name = anno.Index
        else:
            name = anno.name

        if "Polygon" in anno.geometry.geom_type:
            anno_layer = []
            if anno.geometry.geom_type == "Polygon":
                coordinates = anno.geometry.exterior.coords[:]
                anno_layer.append(coordinates)
            elif anno.geometry.geom_type == "MultiPolygon":
                for poly in anno.geometry.geoms:
                    coordinates = poly.exterior.coords[:]
                    anno_layer.append(coordinates)
            else:
                raise NotImplementedError(anno.geometry.geom_type)
            napari_layers.append(
                (
                    anno_layer,
                    {
                        "name": f"Annot. {name}",
                        "shape_type": "polygon",
                        "rotate": -90,
                        "scale": [
                            -1,
                            1,
                        ],  # Needs to be set to img layer scale, with *-1,*1
                        "blending": "translucent",
                        "edge_color": "#55007fff",
                        "edge_width": 60,
                        "face_color": "transparent",
                        "text": {
                            "string": name,
                            "size": 12,
                            "scaling": True,
                            "anchor": "upper_left",
                            "color": "#550000ff",
                        },
                        "metadata": {"qupath_comp": True},
                    },
                    "shapes",
                )
            )
        elif "Point" in anno.geometry.geom_type:
            anno_layer = []
            if anno.geometry.geom_type == "Point":
                coordinates = anno.geometry.coords[0]
                anno_layer.append(coordinates)
            elif anno.geometry.geom_type == "MultiPoint":
                for point in anno.geometry.geoms:
                    coordinates = point.coords[0]
                    anno_layer.append(coordinates)
            else:
                raise NotImplementedError(anno.geometry.geom_type)
            napari_layers.append(
                (
                    anno_layer,
                    {
                        "name": f"Annot. {name}",
                        "rotate": -90,
                        "scale": [
                            -1,
                            1,
                        ],  # Needs to be set to img layer scale, with *-1,*1
                        "text": {
                            "string": name,
                            "size": 12,
                            "scaling": True,
                            "anchor": "upper_left",
                            "color": "#550000ff",
                        },
                        "metadata": {"qupath_comp": True},
                    },
                    "points",
                )
            )
        else:
            print(f"{anno.geometry.geom_type} are not supported!")

    return napari_layers


def write_geojson(out_path: str | pathlib.Path, features: dict) -> list[str]:
    gdf = geopandas.GeoDataFrame.from_features(features)
    # gdf.to_file(filename= pathlib.Path(out_path) / f"anno_{"-".join(saved_annotations)}.geojson", driver="GeoJSON")
    gdf.to_file(filename=f"{pathlib.Path(out_path)}.geojson", driver="GeoJSON")
    return [str(out_path)]


# Parquet#


def load_parquet(path: str | pathlib.Path) -> list[FullLayerData]:
    """
    Load a Parquet file and return its contents as a list of LayerDataTuple.
    Parameters:
    path (str | pathlib.Path): The path to the Parquet file.
    Returns:
    list[LayerDataTuple]: A list containing the data from the Parquet file,
        formatted as a LayerDataTuple with the data as a NumPy array,
        an empty dictionary for metadata, and the string "labels".
    """
    print("WIP function")
    if isinstance(path, str):
        print("PATH IS A STRINGG!!!!!")
        path = pathlib.Path(path)

    df = pandas.read_parquet(path)
    print("Experiental")
    print(df)

    return [(df.to_numpy(), {}, "labels")]
